#ifndef _XFIGDRAWING_H_
#define _XFIGDRAWING_H_

#include "drawing.h"

// si veda /usr/X11R6/lib/X11/xfig/html/fig-format.html

class figdrawer: public drawer {
private:
  ostream &out;
  int left,right,bottom,top; // Bounding box
  int my_linewidth;
  point intrinsic(point p) const {
    return point(double(left)+(right-left)*(p.x-lower_left.x)/
		 (upper_right.x-lower_left.x), 
		 double(top)-(top-bottom)*(p.y-lower_left.y)/
		 (upper_right.y-lower_left.y));};
  void write(point p,ostream &out) const {
    point intr=intrinsic(p);
    out<<int(intr.x*12)<<" "<<int(intr.y*12);
  };
public:
  figdrawer(ostream &o, int l,int r, int b, int t):
    drawer(point(l,b),point(r,t)),
    out(o), left(l), right(r), bottom(b), top(t),
    my_linewidth(1) {
    out<<
      "#FIG 3.2\n"
      "Portrait\n"
      "Center\n"
      "Metric\n"
      "A4\n"
      "100.00\n"
      "Single\n"
      "-2\n"
      "1200 2\n";
  };
  ~figdrawer() {out<<"\n";};

  virtual void graph(const map<double,double> &g) {
    out<<"2 1 0 "<<my_linewidth<<" 0 7 50 0 -1 0.000 0 0 -1 0 0 "<<g.size();
    int count=0;
    for (map<double,double>::const_iterator i=g.begin();i!=g.end();++i) {
      if (count%6==0) out<<"\n\t";
      write(point(i->first,i->second),out);
      out<<" ";
      count++;
    }
    out<<"\n";
  };

  virtual void curve(const vector<point> &g) {
    out<<"2 1 0 "<<my_linewidth<<" 0 3 50 0 20 0.000 0 0 0 0 0 "<<g.size();
    int count=0;
    for (vector<point>::const_iterator i=g.begin();i!=g.end();++i) {
      if (count%5==0) out<<"\n\t";
      write(*i,out);
      out<<" ";
      count++;
    }
    out<<"\n";
  };

  virtual void line(point from, point to) {
    vector<point> g;
    g.push_back(from);
    g.push_back(to);
    curve(g);
  };
  
  virtual void setlinewidth(double w) {
    my_linewidth=int(w*2);
  }
};

#endif
