/*
 * eqdraw: manu-fatto (2002)
 */

#define MYNAME "eqdraw"
#ifndef VERSION
#error VERSION va definito nel Makefile
#endif

#include <iostream>
#include <map>
#include <string>
#include <sstream>
#include <algorithm>
#include <fstream>
#include <memory>
#include <stdexcept>
using namespace std;

#include <ginac/ginac.h>
using namespace GiNaC;

#include "epsdrawing.h"
#include "pngdrawing.h"
#include "figdrawing.h"
#include "eqdraw.h"
#include "options.h"
#include <manucgi.h>

using namespace std;
using namespace manucgi;

class exprFct2: public fct<point,double> {
  const ex E;
  const lst repl;
public:
  exprFct2(const ex &e, const symbol &indep, const symbol &dep): 
    E(e), repl(indep,dep){};
  virtual ~exprFct2(){};
  double operator()(point p) const {
    return ex_to<numeric>(E.subs(repl,lst(p.x,p.y)).evalf()).to_double();
  };
};

class exprFct1: public fct<double,double> {
  const ex E;
  const symbol Indep;
public: 
  exprFct1(const ex &e,const symbol &indep): E(e), Indep(indep) {};
  virtual ~exprFct1(){};
  double operator()(double x) const {
    return ex_to<numeric>(E.subs(Indep==x).evalf()).to_double();
  };
};

class exprFct3: public fct <double,point> {
    const ex X,Y;
  const symbol Param;
public:
  exprFct3(const ex &x, const ex &y ,const symbol &param):
    X(x), Y(y), Param(param) {};
  virtual ~exprFct3(){};
  point operator()(double t) const {
    return point(ex_to<numeric>(X.subs(Param==t).evalf()).to_double(),
		 ex_to<numeric>(Y.subs(Param==t).evalf()).to_double());
  };
};

bool verbose=true;

//string parametric[]={"t","x","y","x0","y0"};
//string *cartesian=parametric+1;

typedef enum {DEFAULT,EPS,PNG,FIG,HTML} output_format_enum;
int output_format=DEFAULT; 
char *output_format_str[]={"eps","png","fig","html",0};
int output_format_vals[]={EPS,PNG,FIG,HTML};

string output_filename="";
bool cgi_mode=false;

int image_width=400, image_height=300;
double xmin=-4, xmax=4, ymin=-3, ymax=3;
double tmin=0, tmax=1;
bool keep_ratio=true;

int eps_page_width=int(sqrt(10000.0/16.0/sqrt(2.0)) /2.54*60.0);
int eps_page_height=int(eps_page_width*sqrt(2.0)); 
int eps_x_off=(eps_page_width-image_width)/2;
int eps_y_off=(eps_page_height-image_height)/2;

Options opt;

char * cgi_options[]={"e","xmin","xmax","ymin","ymax","tmin","tmax",
		      "width","height",0};

class mystream {
  ostream *out;
public:
  mystream(){out=&cout;};
  void reset(ostream *o) {out=o;};
  void away() {
    reset(&cerr);
    if (verbose && !cgi_mode)
      cerr<<"(messages moved from stdout to stderr)\n";
  }
  ostream &operator()(){return *out;};
} out;

string bpar() {
  if (output_format==HTML) return "<p>";
  else return "";
}

string epar() {
  if (output_format==HTML) return "</p>\n";
  else return "\n";
}

string br() {
  if (output_format==HTML) return "<br>\n";
  else return "\n";
}

string bstrong() {
  if (output_format==HTML) return "<strong>";
  else return "";
}

string estrong() {
  if (output_format==HTML) return "</strong>";
  else return "";
}

void write_supported_equations(ostream &out);

void usage(Options &opt) {
  cout<<bpar()<<bstrong()<<MYNAME " ver. " VERSION 
#ifdef DATE
    " (" DATE ") "
#endif
    " - manu-fatto"<<estrong()<<br();
  cout<<"usage:"<<br();
  if (output_format==HTML) cout<<"<pre>\n";
  cout<<MYNAME<<" [options]\n";
  opt.usage(cout);
  cout<<"supported equations:\n";
  write_supported_equations(cout);
  if (output_format==HTML) cout<<"</pre>";
  cout<<epar();
};

class drawer_ptr {
  drawer *draw;
  FILE *fp;
  ostream *o;
  bool prepared;
public:
  drawer_ptr(): draw(0), fp(0), o(o), prepared(false) {};
  void prepare() {
    if (!prepared) open();
    prepared=true;
  };
  void open() {
    close();
    fp=0;
    o=0;
    draw=0;
    if (cgi_mode) output_filename="";
    if (output_format==DEFAULT && output_filename !="") {
      if (verbose)
	out()<<"writing to file "<<output_filename<<"\n";
      string::size_type d=output_filename.rfind('.');
      if (d!=string::npos) {
	string ext=string(output_filename,d+1);
	if (ext=="png" || ext=="PNG") output_format=PNG;
	if (ext=="ps" || ext=="eps" || ext=="PS" || ext=="EPS")
	  output_format=EPS;
	if (ext=="fig" || ext=="FIG" || ext=="xfig")
	  output_format=FIG;
      }
    }
    if (output_format==DEFAULT) output_format=EPS;
    if (output_format==PNG) {
      FILE *f=stdout;
      if (output_filename!="") {
	fp=fopen(output_filename.c_str(),"wb");
	if (!fp) throw runtime_error("Cannot write file "+output_filename);
	f=fp;
      } else {out.away();};
      draw=new pngdrawer(f,image_width,image_height);
    } else if (output_format==EPS || output_format==FIG) {
      ostream *oo=&cout;
      if (output_filename!="") {
	o=new ofstream(output_filename.c_str());
	if (!o || !*o) 
	  throw runtime_error("Cannot write file "+output_filename);
	oo=o;
      } else {out.away();}; 
      if (output_format==FIG) 
	draw=new figdrawer(*oo,eps_x_off,eps_x_off+image_width,
			    eps_y_off,eps_y_off+image_height);
      else
	draw=new epsdrawer(*oo,eps_x_off,eps_x_off+image_width,
			   eps_y_off,eps_y_off+image_height);
    } else if (output_format==HTML) {
      string url=Cgi::GetEnv("SCRIPT_NAME")+string("?");
      for (int i=0;cgi_options[i];++i)
	url+=opt[cgi_options[i]].url();
      cout<<"<img width="<<image_width<<" height="<<image_height
	  <<" src=\""<<url<<"t=png&\"><br>\n"
	"(<a href=\""<<url<<"t=eps&\">versione postscript</a>)\n";
    }
    if (output_format!=HTML) {
      draw->window(point(xmin,ymin),point(xmax,ymax));
      // draw axes 
      draw->setlinewidth(0.5);
      draw->line(point(draw->x0(),0),point(draw->x1(),0));
      draw->line(point(0,draw->y0()),point(0,draw->y1()));
      draw->setlinewidth(1.0);
    }
  };
  void close() {
    delete draw;
    draw=0;
    if (fp) fclose(fp);
    fp=0;
    delete o;
    o=0;
  };
  ~drawer_ptr(){close();};
  drawer* operator->() {prepare();return draw;};
  drawer& operator*() {prepare();return *draw;};
  operator bool() {prepare();return draw!=0;};
} output_draw;

void list_symbols(const ex &e, lst &list) {
  if (is_a<symbol>(e)) {
    list.append(e).sort().unique();
  } else {
    for (unsigned i=0;i<e.nops();++i)
      list_symbols(e.op(i),list);
  }
};

lst list_symbols(const ex &e) {
  lst lista;
  list_symbols(e,lista);
  //  cout<<"Symbols in "<<e<<": "<<lista<<"\n";
  return lista;
}

bool subset(const lst &set1, const lst &set2) {
  for (unsigned i=0;i<set1.nops();++i) {
    unsigned j;
    for (j=0;j<set2.nops() && set1.op(i)!=set2.op(j);++j);
    if (j==set2.nops()) {
      //  cout<<"set "<<set1<<" is not subset of "<<set2<<"\n";
      return false;
    }
  };
  //  cout<<"set "<<set1<<" is subset of "<<set2<<"\n";
  return true;
};



symbol param("t"), indep("x"), dep("y");
symbol dep1("y_");
lst symbols=lst(param,indep,dep,dep1).sort();

void write_supported_equations(ostream &out) {
  out<<dep<<" = f("<<indep<<")\n"
  <<dep1<<" = f("<<indep<<","<<dep<<")\n"
  <<"{f("<<param<<"), "
  <<"g("<<param<<")}\n";
}

void draw(ex F) {
  bool ok=false;
  if (output_format==HTML) output_draw.prepare();
  
  if (F.info(info_flags::relation_equal)) {
    ex left=F.op(0), right=F.op(1);
    lst smbs=list_symbols(right);
    if (left==dep && subset(smbs,lst(indep))) {
      // equazione del tipo y=f(x)
      ok=true;
      if (verbose) 
      out()<<bpar()<<"Grafico di funzione: "<<bstrong()<<F<<estrong()<<epar();
      exprFct1 eF(right,indep);
      if (output_draw) {
	int n=FnDraw(eF,*output_draw);
	if (verbose)
	  out()<<n<<" lines drawn\n";
      }
    } else if (left==dep1 && subset(smbs,lst(dep,indep))) {
      // equazione del tipo y'=f(x,y)
      ok=true;
      if (verbose) { 
	out()<<bpar()<<
	  "Equazione differenziale del prim'ordine in forma normale: "
	     <<bstrong()<<F<<estrong()<<epar();
      }
      exprFct2 eF(right,indep,dep);
      if (output_draw) {
	int n=DiffDrawAll(eF,*output_draw);
	if (verbose) 
	  out()<<n<<" lines drawn\n";
      }
    }
  } else if (F.info(info_flags::list)) {
    lst vec=ex_to<lst>(F);
    if (vec.nops()==2 && subset(list_symbols(vec.op(0)),lst(param))
	&& subset(list_symbols(vec.op(1)),lst(param))) {
      //equazione parametrica
      ok=true;
      if (verbose) 
	out()<<bpar()<<"Supporto della curva: "<<bstrong()
	     <<F<<estrong()<<epar();
      exprFct3 eF(vec.op(0),vec.op(1),param);
      if (output_draw) {
	int n=CurveDraw(eF,*output_draw,tmin,tmax);
	if (verbose)
	  out()<<n<<" lines drawn\n";
      }
      
    }
  }

  if (!ok) {
    out()<<
      "Non so trattare questo tipo di equazioni.\n"
      "Eq: "<<F<<"\n";
    if (verbose) {
      out()<<"Equazioni correntemente supportate:\n";
      write_supported_equations(out());
    }
  }
}

void Optdraw(Options &opt,string s) {
  if (s!="") {
    for (unsigned i=0;i<s.size();++i) {
      if (s[i]=='\'') s[i]='_';
      if (s[i]=='=') {
	if (i+1<s.size() && s[i+1]=='=')
	  i++;
	else
	  s.insert(i++,1,'=');
      }
    }
    //    cout<<"parsing expression: "<<s<<"\n";
    return draw(ex(s,symbols));
  }
};

void set_options() {
  opt.add(Option(new OptNormal("o","output",
			       "-o|--output <filename>: set output file name"),
		 new OptStringRef(output_filename)));
  opt.add(Option(new OptNormal("h","help",
			       "-h|--help: writes help information"),
		 new OptFun(usage,-100)));
  opt.add(Option(new OptNormal("t","type",
				"-t|--type [eps|png|fig|html]: output format"),
		 new OptCases(output_format_str,output_format_vals,
			      output_format)));
  opt.add(Option(new OptNormal("cgi","cgi",
				"-cgi: cgi mode on"),
		 new OptBool(cgi_mode)));
  opt.add(Option(new OptNormal("v","verbose",
				"-v|--verbose: be verbose"),
		  new OptBool(verbose)));

  opt.add(Option(new OptNormal("xmin","xmin",
			       "-xmin <val>"),
		 new OptDoubleRef(xmin)));  
  opt.add(Option(new OptNormal("xmax","xmax",
			       "-xmax <val>"),
		 new OptDoubleRef(xmax)));
  opt.add(Option(new OptNormal("ymin","ymin",
			       "-ymin <val>"),
		 new OptDoubleRef(ymin)));
  opt.add(Option(new OptNormal("ymax","ymax",
			       "-ymax <val>"),
		 new OptDoubleRef(ymax)));
  opt.add(Option(new OptNormal("tmin","tmin",
			       "-tmin <val>"),
		 new OptDoubleRef(tmin)));
  opt.add(Option(new OptNormal("tmax","tmax",
			       "-tmax <val>"),
		 new OptDoubleRef(tmax)));

  opt.add(Option(new OptNormal("width","image-width",
			       "-width|--image-width <val>"),
		 new OptIntRef(image_width)));
  opt.add(Option(new OptNormal("height","image-height",
			       "-height|--image-height <val>"),
		 new OptIntRef(image_height)));

  
  //ultima!
  opt.add(Option(new OptNormal("e","equation","-e|--equation <equation>"),
		 new OptFunForeachString(Optdraw)));

};

Cgi *cgi_ptr=0;

void html_headers() {
  out()<<"<html><head><title>Equation drawer (manu-fatto)</title></head>\n"
    "<body><h1>Equation drawer</h1>\n";
}

void html_footers() {
  if (cgi_ptr && !cgi_ptr->empty()) {
    out()<<"<hr><p><strong>add or modify equations:</strong><br>\n"
	 <<"<form method=\"POST\" action=\""<<Cgi::GetEnv("SCRIPT_NAME")<<"\">\n"
	 <<"Equations: "<<opt["e"].form()
	 <<"<input name=\"e\"><br>\n";
    for (int i=1;cgi_options[i];++i) {
      if (i%2==1) out()<<"<br>";
      out()<<cgi_options[i]<<": "<<opt[cgi_options[i]].form()<<"\n";
    }
    //       <<"<input type=\"reset\" value=\"clear\">\n"
    out()<<"<br><input type=\"submit\" value=\"draw\"><br>\n"
	 <<"</form></p><hr>\n";
  }

  char* esempi[]={"y=x^2/3+sin(5*x)","{t*cos(10*Pi*t), t*cos(6*Pi*t)}","y'=(y-x)*(x^2*y+1)",0};

  out()<<"<p><strong>esempi:</strong>:"
    "<ul>";
  for (int i=0;esempi[i];++i) {
    out()<<"<li> <form method=\"POST\" action=\""
	 <<Cgi::GetEnv("SCRIPT_NAME")<<"\">\n"
	 <<"<input name='e' value=\""<<esempi[i]<<"\">"
	 <<"<input type='submit' value='prova!'></form>\n";
  }
  out()<<"</ul></p>\n";
  out()<<"<p><strong>new drawing:</strong><br>\n"
       <<"<form action=\""<<Cgi::GetEnv("SCRIPT_NAME")<<"\">\n"
       <<"Equation: "
       <<"<input name=\"e\"><br>\n"
    //       <<"<input type=\"reset\" value=\"clear\">\n"
       <<"<input type=\"submit\" value=\"draw\"><br>\n"
       <<"</form></p><hr>\n";

  usage(opt);
  out()<<"</body></html>\n"; 
}

int main(int argc, char *argv[]) {
  try {
    set_options();
    {
      cgi_mode=false;
      string myname=argv[0];
      string::size_type f=myname.rfind('.');
      if (f!=string::npos && string(myname,f) == ".cgi") cgi_mode=true;
      if (getenv("REQUEST_METHOD")!=0) cgi_mode=true;    
    }
    if (cgi_mode) { // sono un cgi!
      output_format=HTML;
      try {
	Cgi the_cgi;
	cgi_ptr=&the_cgi;
	opt.parse(the_cgi);
	if (output_format!=HTML) verbose=false;
      } catch (runtime_error e) {
	cout<<"content-type: text/html\n\n"
	  "<html><body><h1>Cgi Error!</h1>\n"
	    <<e.what()
	    <<"</html>\n";
	throw e;
      }
    } else opt.parse(argc,argv);
    
    if (cgi_mode) {
      output_filename="";
      if (output_format==HTML) {
	cout<<"content-type: text/html\n\n";
      } else if (output_format==PNG) {
	cout<<"content-type: image/png\n\n";
      } else if (output_format==EPS) {
	cout<<"content-type: application/postscript\n\n"; 
      } else if (output_format==FIG) {
	cout<<"content-type: application/x-xfig\n\n";
      }
    }
    if (output_format==HTML) html_headers();
    try {
      opt.exec();
    } catch (runtime_error e) {
      if (cgi_mode && output_format==HTML) 
	cout<<"<br><br><p><strong>Runtime error:</strong> "<<e.what()<<"\n";
      else throw e;
    }
    if (output_format==HTML)
      html_footers();
  } catch (runtime_error e) {
    cerr<<"Runtime error: "<<e.what()<<"\n";
  }
}
